@@ -24,10 +24,12 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <string.h>
 
 #include <config.h>
 #include <glib.h>
 #include <libgnome/gnome-i18n.h>
+#include <libgnome/gnome-config.h>
 #include <gconf/gconf-client.h>
 #include <bonobo.h>
 
@@ -35,13 +37,24 @@
 #include "dictionary.h"
 
 static BonoboObjectClass                  *dictionary_parent_class;
-static POA_GNOME_Spell_Dictionary__vepv    dictionary_vepv;
 
-#define DICT_DEBUG(x) x
+#define DICT_DEBUG(x)
 #define GNOME_SPELL_GCONF_DIR "/GNOME/Spell"
 
 static void release_engines (GNOMESpellDictionary *dict);
 
+static gchar *
+engine_to_language (GNOMESpellDictionary *dict, SpellEngine *se)
+{
+	return (gchar *) se ? g_hash_table_lookup (dict->engines_ht, se) : NULL;
+}
+
+static SpellEngine *
+language_to_engine (GNOMESpellDictionary *dict, gchar *language)
+{
+	return (SpellEngine *) language ? g_hash_table_lookup (dict->engines_ht, language) : NULL;
+}
+
 static void
 raise_error (CORBA_Environment * ev, const gchar *s)
 {
@@ -61,6 +74,8 @@ gnome_spell_dictionary_init (GObject *object)
 
 	dict->changed = TRUE;
 	dict->engines = NULL;
+	dict->languages = g_hash_table_new (g_str_hash, g_str_equal);
+	dict->engines_ht = g_hash_table_new (NULL, NULL);
 }
 
 static void
@@ -69,6 +84,10 @@ dictionary_finalize (GObject *object)
 	GNOMESpellDictionary *dictionary = GNOME_SPELL_DICTIONARY (object);
 
 	release_engines (dictionary);
+	g_hash_table_destroy (dictionary->languages);
+	dictionary->languages = NULL;
+	g_hash_table_destroy (dictionary->engines_ht);
+	dictionary->engines_ht = NULL;
 
 	G_OBJECT_CLASS (dictionary_parent_class)->finalize (object);
 }
@@ -87,6 +106,22 @@ new_engine (const gchar *language)
 	return se;
 }
 
+static gboolean
+remove_language (gpointer key, gpointer value, gpointer user_data)
+{
+	g_free (key);
+
+	return TRUE;
+}
+
+static gboolean
+remove_engine_ht (gpointer key, gpointer value, gpointer user_data)
+{
+	g_free (value);
+
+	return TRUE;
+}
+
 static void
 release_engines (GNOMESpellDictionary *dict)
 {
@@ -101,6 +136,9 @@ release_engines (GNOMESpellDictionary *dict)
 		dict->engines = g_slist_remove (dict->engines, se);
 	}
 
+	g_hash_table_foreach_remove (dict->languages, remove_language, NULL);
+	g_hash_table_foreach_remove (dict->languages, remove_engine_ht, NULL);
+
 	dict->engines = NULL;
 	dict->changed = TRUE;
 }
@@ -111,23 +149,23 @@ static gchar *known_languages [KNOWN_LANGUAGES*2 + 1] = {
 	"ca", N_("Catalan"),
 	"cs", N_("Czech"),
 	"da", N_("Danish"),
-	"de_de", N_("German (Germany)"),
-	"de_ch", N_("German (Swiss)"),
+	"de-DE", N_("German (Germany)"),
+	"de-CH", N_("German (Swiss)"),
 	"el", N_("Greek"),
-	"en_us", N_("English (American)"),
-	"en_gb", N_("English (British)"),
-	"en_ca", N_("English (Canadian)"),
+	"en-US", N_("English (American)"),
+	"en-GB", N_("English (British)"),
+	"en-CA", N_("English (Canadian)"),
 	"eo", N_("Esperanto"),
 	"es", N_("Spanish"),
 	"fo", N_("Faroese"),
-	"fr_fr", N_("French (France)"),
-	"fr_ch", N_("French (Swiss)"),
+	"fr-FR", N_("French (France)"),
+	"fr-CH", N_("French (Swiss)"),
 	"it", N_("Italian"),
 	"nl", N_("Dutch"),
 	"no", N_("Norwegian"),
 	"pl", N_("Polish"),
-	"pt_pt", N_("Portuguese (Portugal)"),
-	"pt_br", N_("Portuguese (Brazilian)"),
+	"pt-PT", N_("Portuguese (Portugal)"),
+	"pt-BR", N_("Portuguese (Brazilian)"),
 	"ru", N_("Russian"),
 	"sk", N_("Slovak"),
 	"sv", N_("Swedish"),
@@ -244,7 +282,7 @@ impl_gnome_spell_dictionary_get_languages (PortableServer_Servant servant, CORBA
 
 		pos = GPOINTER_TO_INT (l->data);
 		seq->_buffer [i].name = CORBA_string_dup (_(known_languages [pos + 1]));
-		seq->_buffer [i].abrev = CORBA_string_dup (known_languages [pos]);
+		seq->_buffer [i].abbreviation = CORBA_string_dup (known_languages [pos]);
 	}
 	CORBA_sequence_set_release (seq, CORBA_TRUE);
 	g_slist_free (langs);
@@ -280,25 +318,19 @@ impl_gnome_spell_dictionary_set_language (PortableServer_Servant  servant,
 		}
 
 		if (len) {
+			SpellEngine *se;
+			
 			one_language = g_strndup (begin, len);
-			dict->engines = g_slist_prepend (dict->engines, new_engine (one_language));
-			g_free (one_language);
+			se = new_engine (one_language);
+			dict->engines = g_slist_prepend (dict->engines, se);
+			g_hash_table_insert (dict->languages, g_strdup (one_language), se);
+			g_hash_table_insert (dict->engines_ht, se, g_strdup (one_language));
 
 			dict->changed = TRUE;
 		}
 	}
 }
 
-static void
-engines_config_replace (GSList *engines, const gchar *key, const gchar *value)
-{
-	for (; engines; engines = engines->next) {
-		SpellEngine *se = (SpellEngine *) engines->data;
-
-		aspell_config_replace (se->config, key, value);
-	}
-}
-
 static void
 update_engine (SpellEngine *se, CORBA_Environment * ev)
 {
@@ -404,45 +436,43 @@ impl_gnome_spell_dictionary_add_word_to_session (PortableServer_Servant servant,
 
 static void
 impl_gnome_spell_dictionary_add_word_to_personal (PortableServer_Servant servant,
-						  const CORBA_char *word, CORBA_Environment *ev)
+						  const CORBA_char *word, const CORBA_char *language, CORBA_Environment *ev)
 {
 	GNOMESpellDictionary *dict = GNOME_SPELL_DICTIONARY (bonobo_object_from_servant (servant));
-	GSList *l;
+	SpellEngine *se;
 
 	g_return_if_fail (word);
 
 	update_engines (dict, ev);
-	DICT_DEBUG (printf ("Dictionary add_word_to_personal: %s\n", word));
-	for (l = dict->engines; l; l = l->next) {
-		SpellEngine *se = (SpellEngine *) l->data;
+	DICT_DEBUG (printf ("Dictionary add_word_to_personal: %s (%s)\n", word, language));
+	se = (SpellEngine *) g_hash_table_lookup (dict->languages, language);
 
-		if (se->speller) {
-			aspell_speller_add_to_personal (se->speller, word, strlen (word));
-			aspell_speller_save_all_word_lists (se->speller);
-		}
+	if (se && se->speller) {
+		aspell_speller_add_to_personal (se->speller, word, strlen (word));
+		aspell_speller_save_all_word_lists (se->speller);
+		DICT_DEBUG (printf ("Added and saved.\n"));
 	}
 }
 
 static void
 impl_gnome_spell_dictionary_set_correction (PortableServer_Servant servant,
-					    const CORBA_char *word, const CORBA_char *replacement, CORBA_Environment *ev)
+					    const CORBA_char *word, const CORBA_char *replacement, const CORBA_char *language, CORBA_Environment *ev)
 {
 	GNOMESpellDictionary *dict = GNOME_SPELL_DICTIONARY (bonobo_object_from_servant (servant));
-	GSList *l;
+	SpellEngine *se;
 
 	g_return_if_fail (word);
 	g_return_if_fail (replacement);
 
 	update_engines (dict, ev);
 	DICT_DEBUG (printf ("Dictionary correction: %s <-- %s\n", word, replacement));
-	for (l = dict->engines; l; l = l->next) {
-		SpellEngine *se = (SpellEngine *) l->data;
+	se = (SpellEngine *) g_hash_table_lookup (dict->languages, language);
 
-		if (se->speller) {
-			aspell_speller_store_replacement (se->speller, word, strlen (word),
-							  replacement, strlen (replacement));
-			aspell_speller_save_all_word_lists (se->speller);
-		}
+	if (se && se->speller) {
+		aspell_speller_store_replacement (se->speller, word, strlen (word),
+						  replacement, strlen (replacement));
+		aspell_speller_save_all_word_lists (se->speller);
+		DICT_DEBUG (printf ("Set and saved.\n"));
 	}
 }
 
@@ -457,7 +487,7 @@ impl_gnome_spell_dictionary_get_suggestions (PortableServer_Servant servant,
 	GSList *l, *suggestion_list = NULL;
 	gint i, len, pos;
 
-	g_return_if_fail (word);
+	g_return_val_if_fail (word, NULL);
 
 	DICT_DEBUG (printf ("Dictionary correction: %s\n", word));
 	update_engines (dict, ev);
@@ -469,7 +499,8 @@ impl_gnome_spell_dictionary_get_suggestions (PortableServer_Servant servant,
 		if (se->speller) {
 			suggestions  = aspell_speller_suggest (se->speller, word, strlen (word));
 			suggestion_list = g_slist_prepend (suggestion_list, (gpointer) suggestions);
-			len += aspell_word_list_size (suggestions);
+			len += 2*aspell_word_list_size (suggestions);
+			suggestion_list = g_slist_prepend (suggestion_list, engine_to_language (dict, se));
 		}
 	}
 
@@ -484,12 +515,18 @@ impl_gnome_spell_dictionary_get_suggestions (PortableServer_Servant servant,
 	pos = 0;
 	for (l = suggestion_list; l; l = l->next) {
 		gint list_len;
+		gchar *language;
 
+		language = (gchar *) l->data;
+		l = l->next;
 		suggestions = (const AspellWordList  *) l->data;
 		elements = aspell_word_list_elements (suggestions);
 		list_len = aspell_word_list_size (suggestions);
-		for (i = 0; i < list_len; i ++, pos ++)
+		for (i = 0; i < list_len; i ++, pos ++) {
 			seq->_buffer [pos] = CORBA_string_dup (aspell_string_enumeration_next (elements));
+			pos ++;
+			seq->_buffer [pos] = CORBA_string_dup (language);
+		}
 		delete_aspell_string_enumeration (elements);
 	}
 	CORBA_sequence_set_release (seq, CORBA_TRUE);
